"""GitHub App/bot configuration."""

from functools import lru_cache
import os
from typing import Optional

import environ
import envparse

# pylint: disable=relative-beyond-top-level
from ..github.config.app import GitHubAppIntegrationConfig
# pylint: disable=relative-beyond-top-level
from .action.config import GitHubActionConfig
# pylint: disable=relative-beyond-top-level
from .runtime.config import RuntimeConfig
# pylint: disable=relative-beyond-top-level
from .server.config import WebServerConfig


@environ.config
class BotAppConfig:
    """Bot app config.

    Construct it as follows::
    >>> from octomachinery.app.config import BotAppConfig
    >>> config = BotAppConfig.from_dotenv()  # for dev env
    >>> config = BotAppConfig.from_env()  # for pure env
    >>>
    """

    github = environ.group(GitHubAppIntegrationConfig)
    action = environ.group(GitHubActionConfig)
    server = environ.group(WebServerConfig)
    runtime = environ.group(RuntimeConfig)

    @classmethod
    @lru_cache(maxsize=1)
    def from_dotenv(
            cls,
            *,
            app_name: Optional[str] = None,
            app_version: Optional[str] = None,
            app_url: Optional[str] = None,
    ):
        """Return an initialized dev config instance.

        Read .env into env vars before that.
        """
        envparse.Env.read_envfile(
            '.env',  # Making it relative to CWD, relative to caller if None
        )
        return cls.from_env(
            app_name=app_name,
            app_version=app_version,
            app_url=app_url,
        )

    @classmethod
    @lru_cache(maxsize=1)
    def from_env(
            cls,
            *,
            app_name: Optional[str] = None,
            app_version: Optional[str] = None,
            app_url: Optional[str] = None,
    ):
        """Return an initialized config instance."""
        env_vars = dict(os.environ)
        if app_name is not None:
            env_vars['OCTOMACHINERY_APP_NAME'] = app_name
        if app_version is not None:
            env_vars['OCTOMACHINERY_APP_VERSION'] = app_version
        if app_url is not None:
            env_vars['OCTOMACHINERY_APP_URL'] = app_url
        return environ.to_config(cls, environ=env_vars)



"""Test for asynchronous operations utility functions."""

import pytest

from octomachinery.utils.asynctools import amap, dict_to_kwargs_cb, try_await


def sync_power2(val):
    """Raise x to the power of 2."""
    return val ** 2


async def async_power2(val):
    """Raise x to the power of 2 asyncronously."""
    return sync_power2(val)


@pytest.mark.parametrize(
    'callback_func',
    (
        sync_power2,
        async_power2,
    ),
)
async def test_amap(callback_func):
    """Test that async map works for both sync and async callables."""
    async def async_iter(*args, **kwargs):
        for _ in range(*args, **kwargs):
            yield _

    test_range = 5

    actual_result = [
        i async for i in amap(callback_func, async_iter(test_range))
    ]
    expected_result = [
        await try_await(callback_func(i)) for i in range(test_range)
    ]
    assert actual_result == expected_result


@pytest.mark.parametrize(
    'callback_func',
    (
        sync_power2,
        async_power2,
    ),
)
async def test_dict_to_kwargs_cb(callback_func):
    """Test that input dict is turned into given (a)sync callable args."""
    test_val = 5
    test_dict = {'val': test_val}

    actual_result = await dict_to_kwargs_cb(callback_func)(test_dict)
    expected_result = await try_await(callback_func(test_val))
    assert actual_result == expected_result


@pytest.mark.parametrize(
    'callback_func,callback_arg',
    (
        (sync_power2, 3),
        (async_power2, 8),
    ),
)
async def test_try_await(callback_func, callback_arg):
    """Test that result is awaited regardless of (a)sync func type."""
    actual_result = await try_await(callback_func(callback_arg))
    expected_result = callback_arg ** 2
    assert actual_result == expected_result


async def test_try_await_bypass_errors():
    """Test that internal callback exceptions are propagated."""
    async def break_callback():
        raise TypeError('It is broken')

    with pytest.raises(TypeError, match='It is broken'):
        await try_await(break_callback())



"""Version tools set."""

import os
from typing import Callable, Optional, Union

from setuptools_scm import get_version


def get_version_from_scm_tag(
        *,
        root: str = '.',
        relative_to: Optional[str] = None,
        local_scheme: Union[Callable, str] = 'node-and-date',
) -> str:
    """Retrieve the version from SCM tag in Git or Hg."""
    try:
        return get_version(
            root=root,
            relative_to=relative_to,
            local_scheme=local_scheme,
        )
    except LookupError:
        return 'unknown'


def cut_local_version_on_upload(version):
    """Return empty local version if uploading to PyPI."""
    is_pypi_upload = os.getenv('PYPI_UPLOAD') == 'true'
    if is_pypi_upload:
        return ''

    # pylint: disable=import-outside-toplevel
    import setuptools_scm.version  # only available during setup time
    return setuptools_scm.version.get_local_node_and_date(version)


def get_self_version():
    """Calculate the version of the dist itself."""
    return get_version_from_scm_tag(local_scheme=cut_local_version_on_upload)



"""Processing outcomes for use from within GitHub Action env."""

import logging

import attr


__all__ = ('ActionSuccess', 'ActionNeutral', 'ActionFailure', )


logger = logging.getLogger(__name__)


@attr.dataclass
class ActionOutcome:  # pylint: disable=too-few-public-methods
    """GitHub Action processing outcome."""

    message: str
    return_code: int

    def raise_it(self):
        """Print the message and exit the program with current code."""
        logger.info(
            'Terminating the GitHub Action processing: %s',
            self.message,
        )
        raise SystemExit(self.return_code)


@attr.dataclass
# pylint: disable=too-few-public-methods
class ActionSuccess(ActionOutcome):
    """GitHub Action successful outcome."""

    return_code: int = attr.ib(default=0, init=False)


@attr.dataclass
# pylint: disable=too-few-public-methods
class ActionFailure(ActionOutcome):
    """GitHub Action failed outcome."""

    return_code: int = attr.ib(default=1)

    @return_code.validator
    def _validate_return_code(  # pylint: disable=no-self-use
            self,
            attribute,  # pylint: disable=unused-argument
            value,
    ):
        if value in NON_FAIL_MODELS:
            raise ValueError(
                f'Return code of `{value}` is illegal to use for failure '
                f'outcome. Use {NON_FAIL_MODELS[value]} instead',
            )


@attr.dataclass
# pylint: disable=too-few-public-methods
class ActionNeutral(ActionOutcome):
    """GitHub Action neutral outcome."""

    # NOTE: It's EX_CONFIG under BSD and EREMCHG under GNU/Linux
    # NOTE: that's why we are using just 78 conventional constant
    # NOTE: here...
    # NOTE: Ref:
    # https://developer.github.com/actions/creating-github-actions\
    # /accessing-the-runtime-environment/#exit-codes-and-statuses
    return_code: int = attr.ib(default=78, init=False)


NON_FAIL_MODELS = {
    0: 'ActionSuccess',
    78: 'ActionNeutral',
}



# -*- coding: utf-8 -*-
"""Configuration of Sphinx documentation generator.

This file does only contain a selection of the most common options. For a
full list see the documentation:
http://www.sphinx-doc.org/en/master/config
"""

from email import message_from_string
from itertools import chain
import pkg_resources


# RTD hack start
def patch_setuptools_in_rtd():
    """Install newer setuptools and octomachinery in RTD."""
    import os  # pylint: disable=import-outside-toplevel
    import sys  # pylint: disable=import-outside-toplevel
    import subprocess  # pylint: disable=import-outside-toplevel

    if not os.getenv('READTHEDOCS') or os.getenv('READTHEDOCS_EXEC'):
        return

    # pylint: disable=unexpected-keyword-arg
    pyenv_python_executable = subprocess.check_output(
        ('pyenv', 'which', 'python3.7'), text=True,
    ).strip()
    setuptools_update_cmd = (
        pyenv_python_executable, '-m',
        'pip', 'install', '--force-reinstall',
        '--cache-dir', '/home/docs/checkouts/readthedocs.org/user_builds'
        '/octomachinery/.cache/pip', 'setuptools >= 40.9.0',
    )
    pip_update_cmd = (
        sys.executable, '-m',
        'pip', 'install', '--force-reinstall',
        '--cache-dir', '/home/docs/checkouts/readthedocs.org/user_builds'
        '/octomachinery/.cache/pip', 'pip >= 19.0.3',
    )
    octomachinery_install_cmd = (
        sys.executable, '-m',
        'pip', 'install', '--force-reinstall',
        '--cache-dir', '/home/docs/checkouts/readthedocs.org/user_builds'
        '/octomachinery/.cache/pip', '..[docs]',
    )
    print('>>>>> Bumping setuptools...', file=sys.stderr)
    subprocess.check_call(setuptools_update_cmd)
    print('>>>>> Bumping pip...', file=sys.stderr)
    subprocess.check_call(pip_update_cmd)
    print('>>>>> Installing octomachinery...', file=sys.stderr)
    subprocess.check_call(octomachinery_install_cmd)

    new_env = dict(os.environ)
    new_env['READTHEDOCS_EXEC'] = 'True'

    print('>>>>> Restarting Sphinx build...', file=sys.stderr)
    print(
        f'Sphinx build command is `{sys.executable} {" ".join(sys.argv)}`',
        file=sys.stderr,
    )
    os.execve(sys.executable, (sys.executable, *sys.argv), new_env)


patch_setuptools_in_rtd()
del patch_setuptools_in_rtd
# RTD hack end

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

def get_supported_pythons(classifiers):
    """Return min and max supported Python version from meta as tuples."""
    py_ver_classifier = 'Programming Language :: Python :: '
    vers = filter(lambda c: c.startswith(py_ver_classifier), classifiers)
    vers = map(lambda c: c[len(py_ver_classifier):], vers)
    vers = filter(lambda c: c[0].isdigit() and '.' in c, vers)
    vers = map(lambda c: tuple(c.split('.')), vers)
    vers = sorted(vers)
    del vers[1:-1]
    if len(vers) < 2:
        vers *= 2
    return vers


def get_github_data(project_urls):
    """Retrieve GitHub user/org and repo name from a bunch of links."""
    partitioned_urls = (p.partition(', ') for p in project_urls)
    for _url_type, _sep, url in partitioned_urls:
        proto, _gh, uri = url.partition('://github.com/')
        if proto not in ('http', 'https'):
            continue
        return uri.split('/')[:2]

    raise LookupError('There are no project URLs pointing to GitHub')


PYTHON_DISTRIBUTION_NAME = 'octomachinery'

PRJ_DIST = pkg_resources.get_distribution(PYTHON_DISTRIBUTION_NAME)
PRJ_PKG_INFO = PRJ_DIST.get_metadata(PRJ_DIST.PKG_INFO)
PRJ_META = message_from_string(PRJ_PKG_INFO)
PRJ_AUTHOR = PRJ_META['Author']
PRJ_LICENSE = PRJ_META['License']
PRJ_SUMMARY = PRJ_META['Summary']
PRJ_DESCRIPTION = PRJ_META['Description']
PRJ_PY_VER_RANGE = get_supported_pythons(PRJ_META.get_all('Classifier'))
PRJ_PY_MIN_SUPPORTED, PRJ_PY_MAX_SUPPORTED = map('.'.join, PRJ_PY_VER_RANGE)
PRJ_GITHUB_USER, PRJ_GITHUB_REPO = get_github_data(
    chain(
        (PRJ_META['Home-page'], ),
        PRJ_META.get_all('Project-URL'),
    ),
)

project = PRJ_DIST.project_name  # pylint: disable=invalid-name
author = PRJ_AUTHOR  # pylint: disable=invalid-name
copyright = f'2019, {author}'  # pylint: disable=invalid-name,redefined-builtin

# The full version, including alpha/beta/rc tags
release = PRJ_DIST.version  # pylint: disable=invalid-name
# The short X.Y version
# pylint: disable=invalid-name
version = pkg_resources.parse_version(release).base_version

rst_epilog = f"""
.. |project| replace:: {project}
.. |min_py_supported| replace:: {PRJ_PY_MIN_SUPPORTED}
.. |max_py_supported| replace:: {PRJ_PY_MAX_SUPPORTED}
"""  # pylint: disable=invalid-name


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = '1.7.5'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinxcontrib.apidoc',
]

# sphinxcontrib.apidoc configuration options
apidoc_extra_args = ['--implicit-namespaces', '../octomachinery']
apidoc_module_dir = '.'
apidoc_output_dir = 'reference'
apidoc_separate_modules = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'description': PRJ_SUMMARY,
    'github_user': PRJ_GITHUB_USER,
    'github_repo': PRJ_GITHUB_REPO,
    'github_type': 'star',
    'github_banner': True,
    'travis_button': True,
    # 'travis_tld': 'com',
    'show_relbars': True,
    'show_related': True,
    'extra_nav_links': {
        'Create a GitHub bot 🤖': 'https://tutorial.octomachinery.dev',
        'octomachinery [www] 🤖': 'https://octomachinery.dev',
    },
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = f'{project}doc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc, f'{project}.tex', f'{project} Documentation',
        'Sviatoslav Sydorenko (@webknjaz)', 'manual',
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc, project, f'{project} Documentation',
        [author], 1,
    ),
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc, project, f'{project} Documentation',
        author, project, 'One line description of project.',
        'Miscellaneous',
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'gidgethub': ('https://gidgethub.readthedocs.io/en/latest/', None),
    'python': ('https://docs.python.org/', None),
    'tutorial': ('https://tutorial.octomachinery.dev/en/latest/', None),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# Patch alabaster theme
# Ref: https://github.com/bitprophet/alabaster/pull/147
# FIXME: drop this hack once the PR is merged & released; pylint: disable=fixme
def set_up_travis_context(
        app, pagename, templatename,  # pylint: disable=unused-argument
        context,
        doctree,  # pylint: disable=unused-argument
):
    """Add complete Travis URLs to Jinja2 context."""
    github_slug = '/'.join(
        # (context['theme_github_user'], context['theme_github_repo']),
        (PRJ_GITHUB_USER, PRJ_GITHUB_REPO),
    )

    travis_button = 'true'  # str(context['theme_travis_button']).lower()
    travis_button_enabled = travis_button == 'true'

    travis_slug = github_slug if travis_button_enabled else travis_button

    travis_tld = 'com'  # context["theme_travis_tld"].strip(".").lower()
    travis_base_uri = 'travis-ci.{}/{}'.format(travis_tld, travis_slug)
    context['theme_travis_build_url'] = 'https://{}'.format(travis_base_uri)
    context['theme_travis_badge_url'] = 'https://api.{}.svg?branch={}'.format(
        travis_base_uri, 'master',  # context['theme_badge_branch'],
    )


def setup(app):
    """Patch the sphinx theme set up stage."""
    app.connect('html-page-context', set_up_travis_context)



"""Octomachinery CLI runner for GitHub Action environments."""

import asyncio
import logging
import typing

from aiohttp.client import ClientSession

# pylint: disable=relative-beyond-top-level
from ...github.entities.action import GitHubAction
# pylint: disable=relative-beyond-top-level
from ...github.errors import GitHubActionError
# pylint: disable=relative-beyond-top-level
from ...github.models.action_outcomes import (
    ActionSuccess, ActionNeutral, ActionFailure,
)
# pylint: disable=relative-beyond-top-level
from ..config import BotAppConfig
# pylint: disable=relative-beyond-top-level
from ..routing.webhooks_dispatcher import (
    route_github_action_event,
)
# pylint: disable=relative-beyond-top-level
from ..runtime.context import RUNTIME_CONTEXT


logger = logging.getLogger(__name__)


async def process_github_action(config):
    """Schedule GitHub Action event for processing."""
    RUNTIME_CONTEXT.config = config  # pylint: disable=assigning-non-slot
    logger.info('Processing GitHub Action event...')

    async with ClientSession() as http_client_session:
        github_action = GitHubAction(
            metadata=config.action,
            http_session=http_client_session,
            user_agent=config.github.user_agent,
        )
        logger.info('GitHub Action=%r', config.action)

        await route_github_action_event(github_action)
    return ActionSuccess('GitHub Action has been processed')


def run(*, config: typing.Optional[BotAppConfig] = None) -> None:
    """Start up a server using CLI args for host and port."""
    if config is None:
        config = BotAppConfig.from_dotenv()

    logging.basicConfig(
        level=logging.DEBUG
        if config.runtime.debug  # pylint: disable=no-member
        else logging.INFO,
    )

    try:
        processing_outcome = asyncio.run(process_github_action(config))
    except GitHubActionError as action_error:
        action_error.terminate_action()
    except KeyboardInterrupt:
        ActionNeutral('Action processing interrupted by user').raise_it()
    except Exception:  # pylint: disable=broad-except
        err_msg = 'Action processing failed unexpectedly'
        logger.exception(err_msg)
        ActionFailure(err_msg).raise_it()
    else:
        processing_outcome.raise_it()



"""GitHub Action wrapper."""

import json
import logging
from uuid import uuid4

from aiohttp.client import ClientSession
import attr
from gidgethub.sansio import Event

# pylint: disable=relative-beyond-top-level
from ...app.action.config import GitHubActionConfig
# pylint: disable=relative-beyond-top-level
from ..api.raw_client import RawGitHubAPI
# pylint: disable=relative-beyond-top-level
from ..api.tokens import GitHubOAuthToken


logger = logging.getLogger(__name__)


@attr.dataclass
class GitHubAction:
    """GitHub Action API wrapper."""

    _metadata: GitHubActionConfig
    """A GitHub Action metadata from envronment vars."""
    _http_session: ClientSession
    """An externally created aiohttp client session."""
    _user_agent: str
    """A User-Agent string to use in HTTP requests to the GitHub API."""

    @property
    def event(self):  # noqa: D401
        """Parsed GitHub Action event data."""
        try:
            # NOTE: This could be async but it probably doesn't matter
            # NOTE: since it's called just once during init and GitHub
            # NOTE: Action runtime only has one event to process
            # pylint: disable=no-member
            with self._metadata.event_path.open() as event_source:
                return Event(
                    json.load(event_source),
                    event=self._metadata.event_name,
                    delivery_id=uuid4(),
                )
        except TypeError:
            return None

    @property
    def token(self):
        """Return GitHub Action access token."""
        return GitHubOAuthToken(self._metadata.token)

    @property
    def api_client(self):  # noqa: D401
        """The GitHub App client."""
        return RawGitHubAPI(
            token=self.token,
            session=self._http_session,
            user_agent=self._user_agent,
        )



"""Exceptions collection related to interactions with GitHub."""

import attr
from gidgethub import GitHubException

# pylint: disable=relative-beyond-top-level
from ..models.action_outcomes import ActionOutcome


class GitHubError(GitHubException):
    """Generic GitHub-related error."""


@attr.dataclass
class GitHubActionError(GitHubError):
    """Generic GitHub-related error."""

    _outcome: ActionOutcome

    def terminate_action(self):
        """Terminate current process using corresponding return code."""
        self._outcome.raise_it()



"""Models representing objects in GitHub Checks API."""

from functools import partial
from typing import List, Optional

import attr


__all__ = ('NewCheckRequest', 'UpdateCheckRequest', 'to_gh_query')


str_attrib = partial(  # pylint: disable=invalid-name
    attr.ib,
    converter=lambda s: str(s) if s is not None else '',
)

int_attrib = partial(attr.ib, converter=int)  # pylint: disable=invalid-name

optional_attrib = partial(  # pylint: disable=invalid-name
    attr.ib,
    default=None,
)

optional_int_attrib = partial(  # pylint: disable=invalid-name
    optional_attrib,
    validator=attr.validators.optional(lambda *_: int(_[-1])),
)

optional_str_attrib = partial(  # pylint: disable=invalid-name
    optional_attrib,
    validator=attr.validators.optional(lambda *_: str(_[-1])),
)

optional_list_attrib = partial(  # pylint: disable=invalid-name
    attr.ib,
    default=[],
    validator=attr.validators.optional(lambda *_: list(_[-1])),
)


def optional_converter(kwargs_dict, convert_to_cls):
    """Instantiate a class instances from dict."""
    if kwargs_dict is not None and not isinstance(kwargs_dict, convert_to_cls):
        return convert_to_cls(**kwargs_dict)
    return kwargs_dict


def optional_list_converter(args_list, convert_to_cls):
    """Convert list items to class instances."""
    if args_list is not None and isinstance(args_list, list):
        return [
            optional_converter(kwargs_dict, convert_to_cls)
            for kwargs_dict in args_list
        ]
    return args_list


@attr.dataclass
class CheckAnnotation:  # pylint: disable=too-few-public-methods
    """Checks API annotation struct."""

    path: str = str_attrib()
    start_line: int = int_attrib()
    end_line: int = int_attrib()
    annotation_level: str = str_attrib(
        validator=attr.validators.in_(
            (
                'notice',
                'warning',
                'failure',
            ),
        ),
    )
    message: str = str_attrib()
    start_column: Optional[int] = optional_int_attrib()
    end_column: Optional[int] = optional_int_attrib()
    title: Optional[str] = optional_str_attrib()
    raw_details: Optional[str] = optional_str_attrib()


@attr.dataclass
class CheckImage:  # pylint: disable=too-few-public-methods
    """Checks API image struct."""

    alt: str = str_attrib()
    image_url: str = str_attrib()
    caption: Optional[str] = optional_str_attrib()


@attr.dataclass
class CheckActions:
    """Checks API actions struct."""

    label: str = str_attrib()
    description: str = str_attrib()
    identifier: str = str_attrib()

    @label.validator
    def label_up_to_20(self, attribute, value):  # pylint: disable=no-self-use
        """Ensure that label is under 20."""
        if len(value) > 20:
            raise ValueError(
                f'`{attribute.name}` must not exceed 20 characters.',
            )

    @description.validator
    def description_up_to_40(self, attribute, value):
        """Ensure that description is under 40."""
        # pylint: disable=no-self-use
        if len(value) > 40:
            raise ValueError(
                f'`{attribute.name}` must not exceed 40 characters.',
            )

    @identifier.validator
    def identifier_up_to_20(self, attribute, value):
        """Ensure that identifier is under 20."""
        # pylint: disable=no-self-use
        if len(value) > 20:
            raise ValueError(
                f'`{attribute.name}` must not exceed 20 characters.',
            )


@attr.dataclass
class CheckOutput:  # pylint: disable=too-few-public-methods
    """Checks API output struct."""

    title: str = str_attrib()
    summary: str = str_attrib()
    text: str = str_attrib(default='')
    annotations: List[CheckAnnotation] = optional_list_attrib(
        converter=partial(
            optional_list_converter,
            convert_to_cls=CheckAnnotation,
        ),
    )
    images: List[CheckImage] = optional_list_attrib(
        converter=partial(optional_list_converter, convert_to_cls=CheckImage),
    )


@attr.dataclass
class BaseCheckRequestMixin:
    """Checks API base check request mixin."""

    name: str = str_attrib()
    details_url: Optional[str] = optional_str_attrib()
    external_id: Optional[str] = optional_str_attrib()
    status: Optional[str] = attr.ib(
        default='queued',
        validator=attr.validators.optional(
            attr.validators.in_(
                (
                    'queued',
                    'in_progress',
                    'completed',
                ),
            ),
        ),
    )
    # '2018-05-27T14:30:33Z', datetime.isoformat():
    started_at: Optional[str] = optional_str_attrib()
    conclusion: Optional[str] = attr.ib(
        # [required] if 'status' is set to 'completed',
        # should be missing if it's unset
        default=None,
        validator=attr.validators.optional(
            attr.validators.in_(
                (
                    'success',
                    'failure',
                    'neutral',
                    'cancelled',
                    'timed_out',
                    'action_required',
                ),
            ),
        ),
    )
    # [required] if 'conclusion' is set  # '2018-05-27T14:30:33Z':
    completed_at: Optional[str] = optional_str_attrib()
    output: Optional[CheckOutput] = optional_attrib(
        converter=partial(optional_converter, convert_to_cls=CheckOutput),
    )
    actions: List[CheckActions] = optional_list_attrib(
        converter=partial(
            optional_list_converter,
            convert_to_cls=CheckActions,
        ),
    )

    @conclusion.validator  # pylint: disable=no-self-use
    def depends_on_status(self, attribute, value):
        """Ensure that conclusion is present if there's status."""
        if self.status == 'completed' and not value:
            raise ValueError(
                f'`{attribute.name}` must be provided if status is completed',
            )

    @completed_at.validator  # pylint: disable=no-self-use
    def depends_on_conclusion(self, attribute, value):
        """Ensure that completed is present if there's conclusion."""
        if self.conclusion is not None and not value:
            raise ValueError(
                f'`{attribute.name}` must be provided '
                'if conclusion is present',
            )

    @actions.validator
    def actions_up_to_3(self, attribute, value):  # pylint: disable=no-self-use
        """Ensure that the number of actions is below 3."""
        if value is not None and len(value) > 3:
            raise ValueError(f'`{attribute.name}` must not exceed 3 items.')


@attr.dataclass
class NewCheckRequestMixin:  # pylint: disable=too-few-public-methods
    """Checks API new check request mixin."""

    head_branch: str = str_attrib()
    head_sha: str = str_attrib()


@attr.dataclass
class NewCheckRequest(NewCheckRequestMixin, BaseCheckRequestMixin):
    """Checks API new check request."""


@attr.dataclass
class UpdateCheckRequest(BaseCheckRequestMixin):
    """Checks API update check request."""


def conditional_to_gh_query(req):
    """Traverse Checks API request structure."""
    if hasattr(req, '__attrs_attrs__'):
        return to_gh_query(req)
    if isinstance(req, list):
        return list(map(conditional_to_gh_query, req))
    if isinstance(req, dict):
        return {
            k: conditional_to_gh_query(v) for k, v in req.items()
            if v is not None or (isinstance(v, (list, dict)) and not v)
        }
    return req


def to_gh_query(req):
    """Convert Checks API request object into a dict."""
    return {
        k: conditional_to_gh_query(v)  # recursive if dataclass or list
        for k, v in attr.asdict(req).items()
        if v is not None or (isinstance(v, (list, dict)) and not v)
    }



"""GitHub webhooks routing."""

from functools import wraps

from gidgethub.routing import Router


__all__ = ('dispatch_event', 'process_event', 'process_event_actions')


WEBHOOK_EVENTS_ROUTER = Router()
"""An event dispatcher for webhooks."""


dispatch_event = WEBHOOK_EVENTS_ROUTER.dispatch  # pylint: disable=invalid-name
process_event = WEBHOOK_EVENTS_ROUTER.register  # pylint: disable=invalid-name


def process_event_actions(event_name, actions=None):
    """Subscribe to multiple events."""
    if actions is None:
        actions = []

    def decorator(original_function):

        def wrapper(*args, **kwargs):
            return original_function(*args, **kwargs)

        if not actions:
            wrapper = process_event(event_name)(wrapper)

        for action in actions:
            wrapper = process_event(event_name, action=action)(wrapper)

        return wraps(original_function)(wrapper)

    return decorator



"""Tests for GitHub models utility functions."""

from datetime import datetime, timezone
from functools import partial

import pytest

from octomachinery.github.models.utils import (
    convert_datetime, SecretStr, SuperSecretStr,
)


# pylint: disable=invalid-name
utc_datetime = partial(datetime, tzinfo=timezone.utc)


@pytest.mark.parametrize(
    'secret_class,secret_placeholder,visible_first_repr',
    (
        (SecretStr, '<SECRET>', True),
        (SuperSecretStr, '<SUPER_SECRET>', False),
    ),
)
def test_secret_sanitizers_first_repr(
        secret_class, secret_placeholder, visible_first_repr,
):
    """Check that immediate repr is rendered correctly."""
    secret_data = 'qwerty'
    super_secret_string = secret_class(secret_data)

    expected_repr = (
        repr(secret_data) if visible_first_repr
        else secret_placeholder
    )
    assert repr(super_secret_string) == expected_repr


@pytest.mark.parametrize(
    'secret_class,secret_placeholder',
    (
        (SecretStr, '<SECRET>'),
        (SuperSecretStr, '<SUPER_SECRET>'),
    ),
)
def test_secret_sanitizers(secret_class, secret_placeholder):
    """Check that sanitizer classes hide data when needed."""
    secret_data = 'qwerty'
    super_secret_string = secret_class(secret_data)
    assert str(super_secret_string) == str(secret_data)
    assert super_secret_string == secret_data

    class _DataStruct:  # pylint: disable=too-few-public-methods
        def __init__(self, s, o=None):
            self._s = s
            self._o = o

        def __repr__(self):
            return (
                f'{self.__class__.__name__}'
                f'(s={repr(self._s)}, o={repr(self._o)})'
            )

    open_data = 'gov'
    data_struct = _DataStruct(super_secret_string, open_data)

    expected_data_struct_repr = (
        f'_DataStruct(s={secret_placeholder}, o={repr(open_data)})'
    )
    assert repr(data_struct) == expected_data_struct_repr

    sub_data_struct = _DataStruct(data_struct, secret_class(None))

    expected_nested_data_struct_repr = (
        f'_DataStruct(s=_DataStruct(s={secret_placeholder}, '
        f'o={repr(open_data)}), o={secret_placeholder})'
    )
    assert repr(sub_data_struct) == expected_nested_data_struct_repr


@pytest.mark.parametrize(
    'input_date_string,expected_date_object',
    (
        (1556634127, utc_datetime(2019, 4, 30, 14, 22, 7)),
        (0, utc_datetime(1970, 1, 1, 0, 0)),
        ('2032-01-02T05:28:47Z', utc_datetime(2032, 1, 2, 5, 28, 47)),
        ('2032-01-02T05:28:47Z+00:00', utc_datetime(2032, 1, 2, 5, 28, 47)),
        (
            '2032-01-02T05:28:47.000Z+00:00',
            utc_datetime(2032, 1, 2, 5, 28, 47),
        ),
        (
            '2032-01-02T05:28:47.000000Z+00:00',
            utc_datetime(2032, 1, 2, 5, 28, 47),
        ),
        ('2032-01-02T05:28:47.000Z', utc_datetime(2032, 1, 2, 5, 28, 47)),
        ('2032-01-02T05:28:47.000000Z', utc_datetime(2032, 1, 2, 5, 28, 47)),
    ),
)
def test_convert_datetime(input_date_string, expected_date_object):
    """Test that convert_datetime recognizes supported date formats."""
    assert convert_datetime(input_date_string) == expected_date_object


@pytest.mark.parametrize(
    'input_date_string',
    (
        None,
        float('NaN'),
        float('Inf'),
        -float('Inf'),
        object(),
        {},
        dict(),
        [],
        frozenset(),
        (),
    ),
)
def test_convert_datetime_negative(input_date_string):
    """Test that convert_datetime errors out on supported date input."""
    with pytest.raises(
            ValueError,
            match=r'^The input arg type .* is not supported$',
    ):
        convert_datetime(input_date_string)


def test_convert_datetime_empty_string():
    """Test that convert_datetime errors out on supported date input."""
    with pytest.raises(
            ValueError,
            match=r'^The input arg .* is unsupported$',
    ):
        convert_datetime('')



"""Models representing objects in GitHub API."""

from datetime import datetime, timezone
import typing

import attr

from .utils import convert_datetime, SecretStr


@attr.dataclass
class GitHubAppInstallation:  # pylint: disable=too-few-public-methods
    """
    Represents a GitHub App installed into a user or an organization profile.

    It has its own ID for installation which is a unique combo of an app
    and a profile (user or org).
    """

    id: int = attr.ib(converter=int)
    """Installation ID."""
    app_id: int = attr.ib(converter=int)
    """GitHub App ID."""

    # FIXME: unignore once this is solved:  # pylint: disable=fixme
    # https://github.com/python/mypy/issues/6172#issuecomment-515718727
    created_at: datetime = attr.ib(converter=convert_datetime)  # type: ignore
    """Date time when the installation has been installed."""
    updated_at: datetime = attr.ib(converter=convert_datetime)  # type: ignore
    """Date time when the installation was last updated."""

    account: typing.Dict[str, typing.Any]
    """Target account (org or user) where this GitHub App is installed into."""
    events: typing.List[str]
    """List of webhook events the app will be receiving from the account."""
    permissions: typing.Dict[str, typing.Any]
    """Permission levels of access to API endpoints types."""
    repository_selection: str = attr.ib(converter=str)
    """Repository selection mode."""
    single_file_name: typing.Optional[str]
    """File path the GitHub app controls."""

    target_id: int = attr.ib(converter=int)
    """Target account ID where this GitHub App is installed into."""
    target_type: str = attr.ib(
        validator=attr.validators.in_(('Organization', 'User')),
    )
    """Target account type where this GitHub App is installed into."""

    access_tokens_url: str = attr.ib(converter=str)
    """API endpoint to retrieve access token from."""
    html_url: str = attr.ib(converter=str)
    """URL for controlling the GitHub App Installation."""
    repositories_url: str = attr.ib(converter=str)
    """API endpoint listing repositories accissible by this Installation."""


@attr.dataclass
class GitHubInstallationAccessToken:  # pylint: disable=too-few-public-methods
    """Struct for installation access token response from GitHub API."""

    token: SecretStr = attr.ib(converter=SecretStr)
    """Access token for GitHub App Installation."""
    expires_at: datetime = attr.ib(converter=convert_datetime)  # type: ignore
    """Token expiration time."""
    permissions: typing.Dict[str, str]
    """Permission levels of access to API endpoints types."""
    repository_selection: str = attr.ib(converter=str)
    """Repository selection mode."""
    repositories: typing.List[typing.Dict[str, typing.Any]] = attr.ib(
        default=[],
        converter=list,
    )
    """List of accessible repositories."""

    @property
    def expired(self):
        """Check whether this token has expired already."""
        return datetime.now(timezone.utc) > self.expires_at



"""Actions processor test suite."""
# pylint: disable=redefined-outer-name

import json

import pytest

from octomachinery.app.action.config import GitHubActionConfig
from octomachinery.app.action.runner import run
from octomachinery.app.config import BotAppConfig
from octomachinery.app.routing import process_event
from octomachinery.app.routing import process_event_actions
from octomachinery.app.runtime.config import RuntimeConfig
from octomachinery.app.server.config import WebServerConfig
from octomachinery.github.config.app import GitHubAppIntegrationConfig
from octomachinery.github.errors import GitHubActionError
from octomachinery.github.models.action_outcomes import ActionNeutral


@process_event('unmatched_event', action='happened')
async def unmatched_event_happened(action):  # pylint: disable=unused-argument
    """Handle an unmached event."""


@process_event('check_run', action='created')
async def check_run_created(action):  # pylint: disable=unused-argument
    """Handle a check_run event."""
    assert False


@process_event_actions('neutral_event', {'qwerty'})
async def neutral_event_qwerty(action):  # pylint: disable=unused-argument
    """Handle a neutral_event."""
    raise GitHubActionError(ActionNeutral('Neutral outcome'))


@process_event_actions('neutral_event')
async def neutral_event_dummy(action):  # pylint: disable=unused-argument
    """Handle any neutral event."""


@pytest.fixture
def event_file(tmp_path_factory, request):
    """Generate a sample JSON event file."""
    gh_event = (
        tmp_path_factory.mktemp('github-action-event') /
        'github-workflow-event.json'
    )
    with gh_event.open('w') as gh_event_file:
        json.dump(
            {
                'action': request.param,  # 'created',
            },
            gh_event_file,
        )
    return gh_event


@pytest.fixture
def config(monkeypatch, event_file, request):
    """Create a dummy GitHub Action config."""
    monkeypatch.setattr(
        'octomachinery.app.runtime.utils.detect_env_mode',
        lambda: 'action',
    )
    return BotAppConfig(
        github=GitHubAppIntegrationConfig(),
        action=GitHubActionConfig(
            workflow='Test Workflow',
            action='Test Action',
            actor='username-or-bot',
            repository='org/repo',
            event_name=request.param,  # 'check_run'
            event_path=str(event_file),  # '/github/workflow/event.json'
            workspace='/github/workspace',
            sha='e6d4abcb8a6cd989d41ee',
            ref='refs/heads/master',
            token='8sdfn12lifds8sdvh832n32f9jew',
        ),
        server=WebServerConfig(),
        runtime=RuntimeConfig(),
    )


@pytest.mark.parametrize(
    'config, event_file, expected_return_code',
    (
        ('check_run', 'created', 1),
        ('unmatched_event', 'closed', 0),
        ('neutral_event', 'qwerty', 78),
    ),
    indirect=('config', 'event_file'),
)
def test_action_processing_return_code(config, expected_return_code):
    """Test an empty action processing run."""
    with pytest.raises(
            SystemExit,
            match=r'^'
            f'{expected_return_code}'
            r'$',
    ):
        run(config=config)



"""Utilitary helpers."""

from functools import wraps
from inspect import Parameter, Signature
from types import AsyncGeneratorType
from typing import Any, Dict, Tuple

from gidgethub.sansio import accept_format


def mark_uninitialized_in_repr(cls):
    """Patch __repr__ for uninitialized instances."""
    orig_repr = cls.__repr__

    @wraps(orig_repr)
    def new_repr(self):
        if not self.is_initialized:
            return f'{self.__class__.__name__}(<UNINITIALIZED>)'
        return orig_repr(self)
    cls.__repr__ = new_repr
    return cls


def accept_preview_version(wrapped_coroutine):
    """Extend keyword-args with `preview_api_version`."""
    @wraps(wrapped_coroutine)
    def coroutine_wrapper(
            *args: Tuple[Any], **kwargs: Dict[str, Any]
    ) -> AsyncGeneratorType:
        accept_media = kwargs.pop('accept', None)
        preview_api_version = kwargs.pop('preview_api_version', None)

        if preview_api_version is not None:
            accept_media = accept_format(
                version=f'{preview_api_version}-preview',
            )
        if accept_media is not None:
            kwargs['accept'] = accept_media

        coroutine_instance = wrapped_coroutine(*args, **kwargs)
        is_async_generator = isinstance(coroutine_instance, AsyncGeneratorType)

        if not is_async_generator:
            async def async_function_wrapper():
                return await coroutine_instance
            return async_function_wrapper()

        async def async_generator_wrapper():
            async for result_item in coroutine_instance:
                yield result_item

        return async_generator_wrapper()

    original_wrapped_signature = Signature.from_callable(wrapped_coroutine)
    original_callable_params = original_wrapped_signature.parameters
    wrapped_callable_params = list(original_callable_params.values())

    accept_pos = list(original_callable_params.keys()).index('accept')
    preview_param = Parameter(
        name='preview_api_version',
        annotation='Optional[str]',
        default=None,
        kind=Parameter.KEYWORD_ONLY,
    )
    wrapped_callable_params.insert(accept_pos, preview_param)

    coroutine_wrapper.__signature__ = original_wrapped_signature.replace(
        parameters=wrapped_callable_params,
    )

    return coroutine_wrapper



"""Config schema for a GitHub App instance details."""
import environ


# pylint: disable=relative-beyond-top-level
from ..models.private_key import GitHubPrivateKey
# pylint: disable=relative-beyond-top-level
from ..models.utils import SecretStr


def validate_is_not_none_if_app(
        self,  # pylint: disable=unused-argument
        attr, value,
):
    """Forbid None value in a GitHub App context."""
    # pylint: disable=relative-beyond-top-level,import-outside-toplevel
    from ...app.runtime.utils import detect_env_mode

    if value is None and detect_env_mode() == 'app':
        raise ValueError(
            f'GitHub App must provide a proper value for {attr!r}',
        )


def validate_fingerprint_if_present(instance, _attribute, value):
    r"""Validate that the private key matches the fingerprint pin.

    :raises ValueError: if the fingerprint pin is present \
                        but doesn't match the private key
    """
    if not value:
        return

    if instance.private_key.matches_fingerprint(value):
        return

    raise ValueError(
        'The private key provided (with a fingerprint of '
        f'{instance.private_key.fingerprint!s}) does not match '
        f'the pinned fingerprint value of {value!s}',
    )


@environ.config
class GitHubAppIntegrationConfig:  # pylint: disable=too-few-public-methods
    """GitHub App auth related config."""

    app_id = environ.var(
        None,
        name='GITHUB_APP_IDENTIFIER',
        validator=validate_is_not_none_if_app,
    )
    private_key = environ.var(
        None,
        name='GITHUB_PRIVATE_KEY',
        converter=lambda raw_data:
        None if raw_data is None else GitHubPrivateKey(raw_data.encode()),
        validator=validate_is_not_none_if_app,
    )
    private_key_fingerprint = environ.var(
        None,
        name='GITHUB_PRIVATE_KEY_FINGERPRINT',
        validator=validate_fingerprint_if_present,
    )
    webhook_secret = environ.var(
        None, name='GITHUB_WEBHOOK_SECRET',
        converter=lambda s: SecretStr(s) if s is not None else s,
    )

    app_name = environ.var(None, name='OCTOMACHINERY_APP_NAME')
    app_version = environ.var(None, name='OCTOMACHINERY_APP_VERSION')
    app_url = environ.var(None, name='OCTOMACHINERY_APP_URL')

    @property
    def user_agent(self):  # noqa: D401
        """The User-Agent value to use when hitting GitHub API."""
        return f'{self.app_name}/{self.app_version} (+{self.app_url})'



"""Tests for GitHub private key class."""
# pylint: disable=redefined-outer-name
from datetime import date
from pathlib import Path
import random
import re

import pytest

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PublicFormat,
    PrivateFormat,
)
from jwt import decode as parse_jwt

from octomachinery.github.models.private_key import GitHubPrivateKey


@pytest.fixture
def rsa_private_key():
    """Generate an RSA private key."""
    return generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend(),
    )


@pytest.fixture
def rsa_public_key(rsa_private_key):
    """Extract a public key out of private one."""
    return rsa_private_key.public_key()


@pytest.fixture
def rsa_private_key_bytes(rsa_private_key) -> bytes:
    r"""Generate an unencrypted PKCS#1 formatted RSA private key.

    Encoded as PEM-bytes.

    This is what the GitHub-downloaded PEM files contain.

    Ref: https://developer.github.com/apps/building-github-apps/\
         authenticating-with-github-apps/
    """
    return rsa_private_key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.TraditionalOpenSSL,  # A.K.A. PKCS#1
        encryption_algorithm=NoEncryption(),
    )


@pytest.fixture
def rsa_public_key_bytes(rsa_public_key) -> bytes:
    """Return a PKCS#1 formatted RSA public key encoded as PEM."""
    return rsa_public_key.public_bytes(
        encoding=Encoding.PEM,
        format=PublicFormat.PKCS1,
    )


@pytest.fixture
def github_private_key(rsa_private_key_bytes: bytes) -> GitHubPrivateKey:
    """Construct a test instance of ``GitHubPrivateKey``."""
    return GitHubPrivateKey(rsa_private_key_bytes)


@pytest.fixture
def rsa_private_key_path(
        rsa_private_key_bytes: bytes,
        tmp_path_factory,
) -> Path:
    """Save the private key to disk as PEM file and provide its path."""
    tmp_dir = tmp_path_factory.mktemp('github_private_key')
    private_key_filename = (
        f'test-github-app.{date.today():%Y-%m-%d}'
        '.private-key.pem'
    )
    private_key_path = tmp_dir / private_key_filename
    private_key_path.write_bytes(rsa_private_key_bytes)
    return private_key_path


def test_github_private_key__from_file(
        github_private_key,
        rsa_private_key_path: Path,
):
    """Test that GitHubPrivateKey from file and bytes are the same."""
    key_from_file = GitHubPrivateKey.from_file(rsa_private_key_path)
    assert key_from_file == github_private_key


def test_github_private_key____repr__(github_private_key):
    """Verify what repr protocol only exposes fingerprint."""
    repr_pattern = re.compile(
        r"^<GitHubPrivateKey\(b_raw_data=b'<SECRET>'\)\s"
        r'with\sSHA\-1\sfingerprint\s'
        r"'[a-f0-9]{2}(:[a-f0-9]{2}){19}'>$",
    )
    assert repr_pattern.match(repr(github_private_key))
    assert repr_pattern.match(f'{github_private_key!r}')


def test_github_private_key____str__(github_private_key):
    """Verify that the string protocol doesn't expose secrets."""
    escaped_private_key_repr = (
        repr(github_private_key).
        replace('(', r'\(').
        replace(')', r'\)')
    )
    exception_message = (
        "<class 'octomachinery.github.models.private_key.GitHubPrivateKey'> "
        'objects do not implement the string protocol '
        'for security reasons. '
        f'The repr of this instance is {escaped_private_key_repr!s}.'
    )
    with pytest.raises(TypeError, match=exception_message):
        str(github_private_key)
    with pytest.raises(TypeError, match=exception_message):
        f'{github_private_key!s}'  # pylint: disable=pointless-statement


def test_github_private_key__make_jwt_for(
        github_private_key: GitHubPrivateKey,
        rsa_public_key_bytes,
):
    """Verify that e2e encoding-decoding of the JWT works."""
    github_app_id = random.randint(0, 9999999)
    jwt_string = github_private_key.make_jwt_for(app_id=github_app_id)
    payload = parse_jwt(
        jwt_string.encode('utf-8'), rsa_public_key_bytes, algorithms='RS256',
    )
    assert payload['iss'] == github_app_id
    assert payload['exp'] - payload['iat'] == 60


def test_github_private_key__make_jwt_for__invalid_timeout(github_private_key):
    """Verify that time offset can't exceed 10 mins."""
    github_app_id = random.randint(0, 9999999)
    with pytest.raises(
            ValueError,
            match='The time offset must be less than 10 minutes',
    ):
        github_private_key.make_jwt_for(app_id=github_app_id, time_offset=601)



"""The globally accessible context vars mapping.

It is supposed to be used as follows:

>>> from octomachinery.app.runtime.context import RUNTIME_CONTEXT

Or shorter:
>>> from octomachinery.app import RUNTIME_CONTEXT
"""
from .utils import _ContextMap


RUNTIME_CONTEXT = _ContextMap(
    app_installation='app installation',
    app_installation_client='app installation client',
    config='config context',
    github_app='github app',
    github_event='GitHub Event',
    IS_GITHUB_ACTION='Is GitHub Action',
    IS_GITHUB_APP='Is GitHub App',
)



"""Web-server constructors."""

import asyncio
import logging

from aiohttp.client import ClientSession
from aiohttp import web

# pylint: disable=relative-beyond-top-level
from ...github.api.app_client import GitHubApp
# pylint: disable=relative-beyond-top-level
from ..routing.webhooks_dispatcher import (
    route_github_webhook_event,
)
# pylint: disable=relative-beyond-top-level
from ..runtime.context import RUNTIME_CONTEXT


logger = logging.getLogger(__name__)


async def start_tcp_site(server_config, aiohttp_server_runner):
    """Return initialized and listening TCP site."""
    host, port = server_config.host, server_config.port
    aiohttp_tcp_site = web.TCPSite(aiohttp_server_runner, host, port)
    await aiohttp_tcp_site.start()
    logger.info(
        f' Serving on http://%s:%s/ '.center(50, '='),
        host, port,
    )
    return aiohttp_tcp_site


async def get_server_runner(http_handler):
    """Initialize server runner."""
    aiohttp_server = web.Server(http_handler)
    aiohttp_server_runner = web.ServerRunner(
        aiohttp_server,
        # handle SIGTERM and SIGINT
        # by raising aiohttp.web_runner.GracefulExit exception
        handle_signals=True,
    )
    await aiohttp_server_runner.setup()
    return aiohttp_server_runner


async def _prepare_github_app(github_app):
    """Set GitHub App in the context."""
    logger.info('Starting the following GitHub App:')
    logger.info(
        '* app id: %s',
        github_app._config.app_id,  # pylint: disable=protected-access
    )
    logger.info(
        '* private key SHA-1 fingerprint: %s',
        # pylint: disable=protected-access
        github_app._config.private_key.fingerprint,
    )
    logger.info(
        '* user agent: %s',
        github_app._config.user_agent,  # pylint: disable=protected-access
    )
    RUNTIME_CONTEXT.github_app = (  # pylint: disable=assigning-non-slot
        github_app
    )


async def _launch_web_server_and_wait_until_it_stops(web_server_config):
    """Start a web server.

    And then block until SIGINT comes in.
    """
    aiohttp_server_runner = await get_server_runner(route_github_webhook_event)
    aiohttp_tcp_site = await start_tcp_site(
        web_server_config,
        aiohttp_server_runner,
    )

    try:
        await asyncio.get_event_loop().create_future()  # block
    except asyncio.CancelledError:
        logger.info(' Stopping the server '.center(50, '='))
        await aiohttp_tcp_site.stop()


async def run_forever(config):
    """Spawn an HTTP server in asyncio context."""
    RUNTIME_CONTEXT.config = config  # pylint: disable=assigning-non-slot
    logger.debug('The GitHub App env is set to `%s`', config.runtime.env)
    async with ClientSession() as aiohttp_client_session:
        github_app = GitHubApp(
            config.github,
            http_session=aiohttp_client_session,
        )
        await github_app.pre_fetch_installs()
        await _prepare_github_app(github_app)
        await _launch_web_server_and_wait_until_it_stops(config.server)



"""Octomachinery CLI runner."""

import asyncio
import logging
import sys
from typing import Optional

from aiohttp.web_runner import GracefulExit
import attr

# pylint: disable=relative-beyond-top-level
from ..config import BotAppConfig
# pylint: disable=relative-beyond-top-level
from .config import WebServerConfig
# pylint: disable=relative-beyond-top-level
from .machinery import run_forever as run_server_forever


logger = logging.getLogger(__name__)


def run(
        *,
        name: Optional[str] = None,
        version: Optional[str] = None,
        url: Optional[str] = None,
        config: Optional[BotAppConfig] = None,
):
    """Start up a server using CLI args for host and port."""
    if (
            config is not None and
            (name is not None or version is not None or url is not None)
    ):
        raise TypeError(
            'run() takes either a BotAppConfig instance as a config argument '
            'or name, version and url arguments.',
        )
    if config is None:
        config = BotAppConfig.from_dotenv(
            app_name=name,
            app_version=version,
            app_url=url,
        )
        if len(sys.argv) > 2:
            config = attr.evolve(
                config,
                server=WebServerConfig(*sys.argv[1:3]),
            )

    logging.basicConfig(
        level=logging.DEBUG
        if config.runtime.debug  # pylint: disable=no-member
        else logging.INFO,
    )
    if config.runtime.debug:  # pylint: disable=no-member
        logger.debug(
            ' App version: %s '.center(50, '='),
            config.github.app_version,
        )

    try:
        asyncio.run(run_server_forever(config))
    except (GracefulExit, KeyboardInterrupt):
        logger.info(' Exiting the app '.center(50, '='))



"""Test suite for version utility helper functions."""
# pylint: disable=redefined-outer-name

import contextlib
import os
import pathlib
import subprocess
import tempfile

import pytest
import setuptools_scm.version

from octomachinery.utils.versiontools import (
    cut_local_version_on_upload,
    get_self_version,
    get_version_from_scm_tag,
)


@contextlib.contextmanager
def working_directory(path):
    """Change working directory and return back to previous on exit."""
    prev_cwd = pathlib.Path.cwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(prev_cwd)


@pytest.fixture
def temporary_working_directory():
    """Create a temporary working directory, cd there and back on exit."""
    with tempfile.TemporaryDirectory() as tmp_git_repo_dir:
        with working_directory(tmp_git_repo_dir) as path:
            yield path


@pytest.fixture
def git_cmd():
    """Provide a Git command helper."""
    git_cmd = ('git', )
    # pylint: disable=unexpected-keyword-arg
    return lambda *args: subprocess.check_output(git_cmd + args, text=True)


@pytest.fixture
def git_init_cmd(git_cmd):
    """Provide a Git init command helper."""
    return lambda *args: git_cmd(*(('init', ) + args))


@pytest.fixture
def git_config_cmd(git_cmd):
    """Provide a Git config command helper."""
    return lambda *args: git_cmd(*(('config', '--local') + args))


@pytest.fixture
def git_commit_cmd(git_cmd):
    """Provide a Git commit command helper."""
    return lambda *args: git_cmd(*(('commit', '--allow-empty') + args))


@pytest.fixture
def git_tag_cmd(git_cmd):
    """Provide a Git tag command helper."""
    return lambda *args: git_cmd(*(('tag', ) + args))


@pytest.fixture
def tmp_git_repo(temporary_working_directory, git_config_cmd, git_init_cmd):
    """Create a temporary Git repo and cd there, coming back upon cleanup."""
    git_init_cmd()
    git_config_cmd('user.name', 'Test User')
    git_config_cmd('user.email', 'test.user@example.com')
    yield temporary_working_directory


def test_get_self_version_in_git_repo(
        monkeypatch,
        tmp_git_repo,  # pylint: disable=unused-argument
        git_cmd, git_commit_cmd, git_tag_cmd,
):
    """Check that get_self_version works properly in existing Git repo."""
    assert get_self_version() == '0.1.dev0'

    git_commit_cmd('-m', 'Test commit')
    git_tag_cmd('v1.3.9')
    assert get_self_version() == '1.3.9'

    git_commit_cmd('-m', 'Test commit 2')
    head_sha1_hash = git_cmd('rev-parse', '--short', 'HEAD').strip()
    assert get_self_version() == f'1.3.10.dev1+g{head_sha1_hash}'

    with monkeypatch.context() as mp_ctx:
        mp_ctx.setenv('PYPI_UPLOAD', 'true')
        assert get_self_version() == f'1.3.10.dev1'


def test_get_self_version_outside_git_repo(
        temporary_working_directory,  # pylint: disable=unused-argument
):
    """Check that version is unknown outside of Git repo."""
    assert get_self_version() == 'unknown'


def test_cut_local_version_on_upload(
        monkeypatch,
        tmp_git_repo,  # pylint: disable=unused-argument
):
    """Test that PEP440 local version isn't emitted when upload."""
    scm_node = 'gfe99188'
    ver = setuptools_scm.version.ScmVersion(
        'v1.1.4',
        distance=3, node='gfe99188',
        dirty=False, branch='master',
    )
    assert cut_local_version_on_upload(ver) == f'+{scm_node}'

    with monkeypatch.context() as mp_ctx:
        mp_ctx.setenv('PYPI_UPLOAD', 'true')
        assert cut_local_version_on_upload(ver) == ''


def test_get_version_from_scm_tag_in_git_repo(
        monkeypatch,
        tmp_git_repo,  # pylint: disable=unused-argument
        git_cmd, git_commit_cmd, git_tag_cmd,
):
    """Check that get_version_from_scm_tag works properly in Git repo."""
    assert get_self_version() == '0.1.dev0'

    git_commit_cmd('-m', 'Test commit')
    git_tag_cmd('v1.3.9')
    assert get_version_from_scm_tag() == '1.3.9'

    git_commit_cmd('-m', 'Test commit 2')
    head_sha1_hash = git_cmd('rev-parse', '--short', 'HEAD').strip()
    assert get_version_from_scm_tag() == f'1.3.10.dev1+g{head_sha1_hash}'

    with monkeypatch.context() as mp_ctx:
        mp_ctx.setenv('PYPI_UPLOAD', 'true')
        assert get_version_from_scm_tag() == f'1.3.10.dev1+g{head_sha1_hash}'


def test_get_version_from_scm_tag_outside_git_repo(
        temporary_working_directory,  # pylint: disable=unused-argument
):
    """Check that version is unknown outside of Git repo."""
    assert get_version_from_scm_tag() == 'unknown'



"""Webhook event processing helper decorators."""

from functools import wraps


def process_webhook_payload(wrapped_function):
    """Bypass event object keys-values as args to the handler."""
    @wraps(wrapped_function)
    def wrapper(event):
        return wrapped_function(**event.data)
    return wrapper



"""GitHub webhook events dispatching logic."""

import asyncio
import contextlib
from functools import wraps
from http import HTTPStatus
import logging
import typing

from aiohttp import web
from gidgethub import BadRequest, ValidationFailure

# pylint: disable=relative-beyond-top-level,import-error
from ..runtime.context import RUNTIME_CONTEXT
from . import dispatch_event


__all__ = ('route_github_action_event', 'route_github_webhook_event', )


logger = logging.getLogger(__name__)


EVENT_LOG_TMPL = (
    'Got a{}valid X-GitHub-Event=%s '
    'with X-GitHub-Delivery=%s '
    'and X-Hub-Signature=%s'
)

EVENT_INVALID_CHUNK = 'n in'
EVENT_VALID_CHUNK = ' '

EVENT_LOG_VALID_MSG = EVENT_LOG_TMPL.format(EVENT_VALID_CHUNK)
EVENT_LOG_INVALID_MSG = EVENT_LOG_TMPL.format(EVENT_INVALID_CHUNK)


async def get_event_from_request(request):
    """Retrieve Event out of HTTP request if it's valid."""
    webhook_event_signature = request.headers.get(
        'X-Hub-Signature', '<MISSING>',
    )
    try:
        event = await RUNTIME_CONTEXT.github_app.event_from_request(request)
    except ValidationFailure as no_signature_exc:
        logger.info(
            EVENT_LOG_INVALID_MSG,
            request.headers.get('X-GitHub-Event'),
            request.headers.get('X-GitHub-Delivery'),
            webhook_event_signature,
        )
        raise web.HTTPForbidden from no_signature_exc
    else:
        logger.info(
            EVENT_LOG_VALID_MSG,
            event.event,
            event.delivery_id,
            webhook_event_signature,
        )
        return event


def validate_allowed_http_methods(*allowed_methods: str):
    """Block disallowed HTTP methods."""
    _allowed_methods: typing.Set[str]
    if not allowed_methods:
        _allowed_methods = {'POST'}
    else:
        _allowed_methods = set(allowed_methods)

    def decorator(wrapped_function):
        @wraps(wrapped_function)
        async def wrapper(request):
            if request.method not in _allowed_methods:
                raise web.HTTPMethodNotAllowed(
                    method=request.method,
                    allowed_methods=_allowed_methods,
                ) from BadRequest(HTTPStatus.METHOD_NOT_ALLOWED)
            return await wrapped_function(request)
        return wrapper
    return decorator


@validate_allowed_http_methods('POST')
async def route_github_webhook_event(request):
    """Dispatch incoming webhook events to corresponsing handlers."""
    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.IS_GITHUB_ACTION = False
    RUNTIME_CONTEXT.IS_GITHUB_APP = True  # pylint: disable=assigning-non-slot

    github_app = RUNTIME_CONTEXT.github_app

    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.github_event = event = (
        await get_event_from_request(request)
    )

    with contextlib.suppress(LookupError):
        # pylint: disable=pointless-string-statement
        """Provision an installation API client if possible.

        Some events (like `ping`) are
        happening application/GitHub-wide and are not bound to
        a specific installation. The webhook payloads of such events
        don't contain any reference to an installaion.
        Some events don't even refer to a GitHub App
        (e.g. `security_advisory`).
        """
        github_install = await github_app.get_installation(event)
        # pylint: disable=assigning-non-slot
        RUNTIME_CONTEXT.app_installation = github_install
        # pylint: disable=assigning-non-slot
        RUNTIME_CONTEXT.app_installation_client = github_install.api_client

    await asyncio.sleep(1)  # Give GitHub a sec to deal w/ eventual consistency
    await dispatch_event(event)
    return web.Response(
        text=f'OK: GitHub event received. It is {event.event!s} ({event!r})',
    )


async def route_github_action_event(github_action):
    """Dispatch a GitHub action event to corresponsing handlers."""
    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.IS_GITHUB_ACTION = True
    RUNTIME_CONTEXT.IS_GITHUB_APP = False  # pylint: disable=assigning-non-slot
    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.github_event = github_action.event

    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.app_installation_client = github_action.api_client
    await dispatch_event(github_action.event)



"""GitHub token types definitions."""

import attr

# pylint: disable=relative-beyond-top-level
from ..models.utils import SecretStr


@attr.dataclass
class GitHubToken:  # pylint: disable=too-few-public-methods
    """Base class for GitHub tokens."""

    _token_value: SecretStr = attr.ib(
        lambda s: SecretStr(s) if s is not None else s,
    )

    def __str__(self):
        """Render the token as its string value."""
        return self._token_value


@attr.dataclass
# pylint: disable=too-few-public-methods
class GitHubOAuthToken(GitHubToken):
    r"""GitHub OAuth Token.

    It can represent either App Installation token or a personal one.

    Ref: https://developer.github.com\
         /apps/building-github-apps/authenticating-with-github-apps\
         /#authenticating-as-an-installation
    """


@attr.dataclass
# pylint: disable=too-few-public-methods
class GitHubJWTToken(GitHubToken):
    r"""GitHub JSON Web Token.

    It represents GitHub App token.

    Ref: https://developer.github.com\
         /apps/building-github-apps\
         /authenticating-with-github-apps/#authenticating-as-a-github-app
    """



"""Utility helpers for App/Action installations."""

from base64 import b64decode
from http import HTTPStatus
from io import StringIO
from pathlib import Path
import typing

import gidgethub
import yaml

from .context import RUNTIME_CONTEXT


def _get_file_contents_from_fs(file_name: str) -> typing.Optional[str]:
    """Read file contents from file system checkout.

    This code path is synchronous.

    It doesn't matter much in GitHub Actions
    but can be refactored later.
    """
    config_path = (Path('.') / file_name)

    try:
        return config_path.read_text()
    except FileNotFoundError:
        return None


async def _get_file_contents_from_api(
        file_name: str,
        ref: typing.Optional[str],
) -> typing.Optional[str]:
    """Read file contents using GitHub API."""
    github_api = RUNTIME_CONTEXT.app_installation_client
    repo_slug = RUNTIME_CONTEXT.github_event.data['repository']['full_name']

    api_query_params = f'?ref={ref}' if ref else ''
    try:
        config_response = await github_api.getitem(
            f'/repos/{repo_slug}/contents'
            f'/{file_name}{api_query_params}',
        )
    except gidgethub.BadRequest as http_bad_req:
        if http_bad_req.status_code == HTTPStatus.NOT_FOUND:
            return None

        raise

    config_file_found = (
        config_response.get('encoding') == 'base64' and
        'content' in config_response
    )
    if not config_file_found:
        return None

    return b64decode(config_response['content']).decode()


async def read_file_contents_from_repo(
        *,
        file_path: str,
        ref: typing.Optional[str] = None,
) -> typing.Optional[str]:
    """Get a config object from the current installation.

    Read from file system checkout in case of GitHub Action env.
    Grab it via GitHub API otherwise.

    Usage::

        >>> from octomachinery.app.runtime.installation_utils import (
        ...     read_file_contents_from_repo
        ... )
        >>> await read_file_contents_from_repo(
        ...     '/file/path.txt',
        ...     ref='bdeaf38',
        ... )
    """
    if RUNTIME_CONTEXT.IS_GITHUB_ACTION and ref is None:
        return _get_file_contents_from_fs(file_path)

    return await _get_file_contents_from_api(file_path, ref)


async def get_installation_config(
        *,
        config_name: str = 'config.yml',
        ref: typing.Optional[str] = None,
) -> typing.Mapping[str, typing.Any]:
    """Get a config object from the current installation.

    Read from file system checkout in case of GitHub Action env.
    Grab it via GitHub API otherwise.

    Usage::

        >>> from octomachinery.app.runtime.installation_utils import (
        ...     get_installation_config
        ... )
        >>> await get_installation_config()
    """
    config_path = f'.github/{config_name}'

    config_content = await read_file_contents_from_repo(
        file_path=config_path,
        ref=ref,
    )

    if config_content is None:
        return {}

    return yaml.load(StringIO(config_content))



"""GitHub Action environment and metadata representation."""

from pathlib import Path

import environ

# pylint: disable=relative-beyond-top-level
from ...github.models.utils import SecretStr


@environ.config
class GitHubActionConfig:  # pylint: disable=too-few-public-methods
    """GitHub Action config."""

    workflow = environ.var(
        None, name='GITHUB_WORKFLOW',
    )
    action = environ.var(
        None, name='GITHUB_ACTION',
    )
    actor = environ.var(
        None, name='GITHUB_ACTOR',
    )
    repository = environ.var(
        None, name='GITHUB_REPOSITORY',
    )
    event_name = environ.var(
        None, name='GITHUB_EVENT_NAME',
    )
    event_path = environ.var(
        None, converter=lambda p: p if p is None else Path(p),
        name='GITHUB_EVENT_PATH',
    )
    workspace = environ.var(
        None, name='GITHUB_WORKSPACE',
    )
    sha = environ.var(
        None, name='GITHUB_SHA',
    )
    ref = environ.var(
        None, name='GITHUB_REF',
    )
    token = environ.var(
        None, name='GITHUB_TOKEN',
        converter=lambda t: t if t is None else SecretStr(t),
    )



"""A collection of utility functions helping with models."""

from datetime import datetime, timezone
from functools import singledispatch
import sys


@singledispatch
def convert_datetime(datetime_obj) -> datetime:
    """Convert arbitrary object into a datetime instance."""
    raise ValueError(
        f'The input arg type {type(datetime_obj)} is not supported',
    )


@convert_datetime.register
def _convert_datetime_from_unixtime(date_unixtime: int) -> datetime:
    return datetime.fromtimestamp(date_unixtime, timezone.utc)


@convert_datetime.register
def _convert_datetime_from_string(date_string: str) -> datetime:
    if not date_string:
        raise ValueError(
            f'The input arg {date_string!r} is unsupported',
        )
    date_string = date_string.replace('.000Z', '.000000Z')
    if '.' not in date_string:
        date_string = date_string.replace('Z', '.000000Z')
    if '+' not in date_string:
        date_string += '+00:00'

    # datetime.fromisoformat() doesn't understand microseconds
    return datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%fZ%z')


class SecretStr(str):
    """String that censors its __repr__ if called from another repr."""

    def __repr__(self):
        """Produce a string representation."""
        frame_depth = 1

        try:
            while True:
                frame = sys._getframe(  # pylint: disable=protected-access
                    frame_depth,
                )
                frame_depth += 1

                if frame.f_code.co_name == '__repr__':
                    return '<SECRET>'
        except ValueError:
            pass

        return super().__repr__()


class SuperSecretStr(SecretStr):
    """String that always censors its __repr__."""

    def __repr__(self):
        """Produce a string representation."""
        return '<SUPER_SECRET>'



"""GitHub App Installation wrapper."""

from __future__ import annotations

import logging
import typing

import attr

# pylint: disable=relative-beyond-top-level
from ..api.raw_client import RawGitHubAPI
# pylint: disable=relative-beyond-top-level
from ..api.tokens import GitHubOAuthToken
# pylint: disable=relative-beyond-top-level
from ..models import (
    GitHubAppInstallation as GitHubAppInstallationModel,
    GitHubInstallationAccessToken,
)


if typing.TYPE_CHECKING:
    from ..api.app_client import GitHubApp


logger = logging.getLogger(__name__)


@attr.dataclass
class GitHubAppInstallation:
    """GitHub App Installation API wrapper."""

    _metadata: GitHubAppInstallationModel
    """A GitHub Installation metadata from GitHub webhook."""
    _github_app: GitHubApp
    """A GitHub App the Installation is associated with."""
    _token: GitHubInstallationAccessToken = attr.ib(init=False, default=None)
    """A GitHub Installation token for GitHub API."""

    @property
    def app(self):
        """Bound GitHub App instance."""
        return self._github_app

    async def get_token(self):
        """Retrieve installation access token from GitHub API."""
        return GitHubInstallationAccessToken(**(
            await self.app.api_client.post(
                self._metadata.access_tokens_url,
                data=b'',
                preview_api_version='machine-man',
            )
        ))

    async def _refresh_api_token(self):
        """Extract installation access token value.

        Refreshes it as needed.
        """
        if self._token is None or self._token.expired:
            self._token = await self.get_token()

        return GitHubOAuthToken(self._token.token)

    @property
    def api_client(self):  # noqa: D401
        """The GitHub App Installation client."""
        return RawGitHubAPI(
            token=self._refresh_api_token,
            # pylint: disable=protected-access
            session=self.app._http_session,
            # pylint: disable=protected-access
            user_agent=self.app._config.user_agent,
        )



"""Private key container."""
from hashlib import sha1 as compute_sha1_hash
from pathlib import Path
from time import time

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from jwt import encode as compute_jwt


def extract_private_key_sha1_fingerprint(rsa_private_key):
    r"""Retrieve the private key SHA-1 fingerprint.

    :param rsa_private_key: private key object
    :type rsa_private_key: cryptography.hazmat.primitives.asymmetric.\
                           rsa.RSAPrivateKey

    :returns: colon-separated SHA-1 fingerprint
    :rtype: str
    """
    rsa_public_key = rsa_private_key.public_key()
    b_rsa_public_key = rsa_public_key.public_bytes(
        Encoding.DER,
        PublicFormat.SubjectPublicKeyInfo,
    )
    rsa_public_key_sha1_fingerprint = compute_sha1_hash(
        b_rsa_public_key,
    ).hexdigest()

    def emit_chunks(sequence, step):
        start_pos = 0
        seq_length = len(sequence)
        while start_pos < seq_length:
            end_pos = start_pos + step
            yield sequence[start_pos: end_pos]
            start_pos = end_pos

    return ':'.join(
        emit_chunks(rsa_public_key_sha1_fingerprint, 2),
    )


class GitHubPrivateKey:
    """Private key entity with a pre-calculated SHA-1 fingerprint.

    :param bytes b_raw_data: the contents of a PEM file
    """

    def __init__(self, b_raw_data: bytes):
        """Initialize GitHubPrivateKey instance."""
        self._rsa_private_key = load_pem_private_key(
            b_raw_data,
            password=None,
            backend=default_backend(),
        )
        self._col_separated_rsa_public_key_sha1_fingerprint = (
            extract_private_key_sha1_fingerprint(self._rsa_private_key)
        )

    @property
    def fingerprint(self) -> str:
        """Colon-separated SHA-1 fingerprint string value.

        :returns: colon-separated SHA-1 fingerprint
        :rtype: str
        """
        return self._col_separated_rsa_public_key_sha1_fingerprint

    def __str__(self):
        """Avoid leaking private key contents via string protocol.

        :raises TypeError: always
        """
        raise TypeError(
            f'{type(self)} objects do not implement the string protocol '
            'for security reasons. '
            f'The repr of this instance is {self!r}.',
        )

    def __repr__(self):
        r"""Construct a GitHubPrivateKey object representation.

        :returns: GitHubPrivateKey object representation \
                  with its SHA-1 fingerprint
        :rtype: str
        """
        return (
            "<GitHubPrivateKey(b_raw_data=b'<SECRET>') "
            f"with SHA-1 fingerprint '{self.fingerprint}'>"
        )

    def __eq__(self, other_private_key):
        r"""Compare equality of our private key with other.

        :returns: the result of comparison with another \
                  ``GitHubPrivateKey`` instance
        :rtype: bool
        """
        return self.matches_fingerprint(other_private_key.fingerprint)

    def matches_fingerprint(self, other_hash):
        """Compare our SHA-1 fingerprint with ``other_hash``.

        :returns: the result of own fingerprint comparison with ``other_hash``
        :rtype: bool
        """
        return self.fingerprint == other_hash

    @classmethod
    def from_file(cls, path):
        r"""Construct a ``GitHubPrivateKey`` instance.

        :returns: the ``GitHubPrivateKey`` instance \
                  constructed of the target file contents
        :rtype: GitHubPrivateKey
        """
        return cls(Path(path).expanduser().read_bytes())

    def make_jwt_for(self, *, app_id: int, time_offset: int = 60) -> str:
        r"""Generate app's JSON Web Token.

        :param int app_id: numeric ID of a GitHub App
        :param int time_offset: duration of the JWT's validity, in seconds, \
                                defaults to 60

        :returns: JWT string for a GitHub App valid for the given time
        :rtype: str

        :raises ValueError: if time_offset exceeds 600 seconds (10 minutes)
        """
        ten_min = 60 * 10
        if time_offset > ten_min:
            raise ValueError('The time offset must be less than 10 minutes')

        now = int(time())
        payload = {
            'iat': now,
            'exp': now + time_offset,
            'iss': app_id,
        }

        return compute_jwt(
            payload,
            key=self._rsa_private_key,
            algorithm='RS256',
        ).decode('utf-8')



"""Test event routing decorator helpers."""

from gidgethub.sansio import Event
import pytest

from octomachinery.app.routing.decorators import process_webhook_payload


@process_webhook_payload
def fake_event_handler(*, arg1, arg2):
    """Process fake test event."""
    return arg1, arg2


@pytest.mark.parametrize(
    'incoming_event,is_successful',
    (
        ({'arg1': 'y', 'arg2': 'x'}, True),
        ({'arg0': 'p', 'arg1': 'u', 'arg2': 'n'}, False),
        ({'arg1': 'z'}, False),
        ({'arg3': 's'}, False),
        (dict(), False),
    ),
)
def test_process_webhook_payload(incoming_event, is_successful):
    """Test that @process_webhook_payload unpacks event into kw-args."""
    event = Event(data=incoming_event, event=None, delivery_id=None)

    if is_successful:
        assert (
            # pylint: disable=missing-kwoa,too-many-function-args
            fake_event_handler(event)
            == tuple(incoming_event.values())
        )
    else:
        with pytest.raises(TypeError):
            # pylint: disable=missing-kwoa,too-many-function-args
            fake_event_handler(event)



"""GitHub App API client."""

from collections import defaultdict
import logging

from aiohttp.client import ClientSession
from aiohttp.client_exceptions import ClientConnectorError
import attr
from gidgethub.sansio import Event

# pylint: disable=relative-beyond-top-level
from ...utils.asynctools import (
    amap, dict_to_kwargs_cb,
)
# pylint: disable=relative-beyond-top-level
from ..config.app import GitHubAppIntegrationConfig
# pylint: disable=relative-beyond-top-level
from ..entities.app_installation import GitHubAppInstallation
# pylint: disable=relative-beyond-top-level
from ..models import GitHubAppInstallation as GitHubAppInstallationModel
from .raw_client import RawGitHubAPI
from .tokens import GitHubJWTToken


logger = logging.getLogger(__name__)


GH_INSTALL_EVENTS = {'integration_installation', 'installation'}


@attr.dataclass
class GitHubApp:
    """GitHub API wrapper."""

    _config: GitHubAppIntegrationConfig
    _http_session: ClientSession

    def __attrs_post_init__(self):
        """Initialize installations store."""
        webhook_secret = self._config.webhook_secret
        webhook_secret_repr = (
            f' ({webhook_secret[:1]}...{webhook_secret[-1:]})'
            if webhook_secret else ''
        )
        logger.info(
            'Webhook secret%s is %sSET.%s',
            webhook_secret_repr,
            '' if webhook_secret else 'NOT ',
            ' SIGNATURE VERIFICATION WILL BE ENFORCED'
            if webhook_secret else '',
        )
        # pylint: disable=attribute-defined-outside-init
        self._installations = defaultdict(dict)

    async def event_from_request(self, request):
        """Get an event object out of HTTP request."""
        event = Event.from_http(
            request.headers,
            await request.read(),
            secret=self._config.webhook_secret,
        )
        await self.pre_process_webhook_event(event)
        return event

    async def pre_process_webhook_event(self, event):
        """Get an event object out of HTTP request."""
        action = event.data.get('action')
        if event.event in GH_INSTALL_EVENTS and action == 'created':
            await self.add_installation(event)

    async def pre_fetch_installs(self) -> 'GitHubApp':
        """Store all installations data before starting."""
        # pylint: disable=attribute-defined-outside-init
        try:
            self._installations = await self.get_installations()
        except ClientConnectorError as client_error:
            logger.info('It looks like the GitHub API is offline...')
            logger.error(
                'The following error has happened while trying to grab '
                'installations list: %s',
                client_error,
            )
            self._installations = defaultdict(dict)

        logger.info('This GitHub App is installed into:')
        # pylint: disable=protected-access
        for install_id, install_val in self._installations.items():
            logger.info(
                '* Installation id %s (installed to %s)',
                install_id,
                install_val._metadata.account['login'],
            )

        return self

    @property
    def gh_jwt(self):
        """Generate app's JSON Web Token, valid for 60 seconds."""
        token = self._config.private_key.make_jwt_for(
            app_id=self._config.app_id,
        )
        return GitHubJWTToken(token)

    @property
    def api_client(self):  # noqa: D401
        """The GitHub App client with an async CM interface."""
        return RawGitHubAPI(
            token=self.gh_jwt,
            session=self._http_session,
            user_agent=self._config.user_agent,
        )

    async def add_installation(self, event):
        """Retrieve an installation creds from store."""
        install = event.data['installation']
        install_id = install['id']
        self._installations[install_id] = GitHubAppInstallation(
            GitHubAppInstallationModel(**install),
            self,
        )
        return self._installations[install_id]

    async def get_installation(self, event):
        """Retrieve an installation creds from store."""
        if 'installation' not in event.data:
            raise LookupError('This event occured outside of an installation')

        install_id = event.data['installation']['id']
        return await self.get_installation_by_id(install_id)

    async def get_installation_by_id(self, install_id):
        """Retrieve an installation with access tokens via API."""
        return GitHubAppInstallation(
            GitHubAppInstallationModel(
                **(await self.api_client.getitem(
                    '/app/installations/{installation_id}',
                    url_vars={'installation_id': install_id},
                    preview_api_version='machine-man',
                )),
            ),
            self,
        )

    async def get_installations(self):
        """Retrieve all installations with access tokens via API."""
        installations = defaultdict(dict)
        async for install in amap(
                dict_to_kwargs_cb(GitHubAppInstallationModel),
                self.api_client.getiter(
                    '/app/installations',
                    preview_api_version='machine-man',
                ),
        ):
            installations[install.id] = GitHubAppInstallation(
                install, self,
            )
        return installations



"""The application runtime configuration."""
import attr
import environ

from .utils import detect_env_mode


@environ.config
class RuntimeConfig:  # pylint: disable=too-few-public-methods
    """Config of runtime env."""

    debug = environ.bool_var(False, name='DEBUG')
    env = environ.var(
        'prod', name='ENV',
        validator=attr.validators.in_(('dev', 'prod')),
    )
    mode = environ.var(
        'auto', name='OCTOMACHINERY_APP_MODE',
        converter=lambda val: detect_env_mode() if val == 'auto' else val,
        validator=attr.validators.in_(('app', 'action')),
    )



"""Asynchronous tools set."""


async def try_await(potentially_awaitable):
    """Try awaiting the arg and return it regardless."""
    valid_exc_str = (
        "can't be used in 'await' expression"
    )

    try:
        return await potentially_awaitable
    except TypeError as type_err:
        type_err_msg = str(type_err)
        if not (
                type_err_msg.startswith('object ')
                and type_err_msg.endswith(valid_exc_str)
        ):
            raise

    return potentially_awaitable


async def amap(callback, async_iterable):
    """Map asyncronous generator with a coroutine or a function."""
    async for async_value in async_iterable:
        yield await try_await(callback(async_value))


def dict_to_kwargs_cb(callback):
    """Return a callback mapping dict to keyword arguments."""
    async def callback_wrapper(args_dict):
        return await try_await(callback(**args_dict))
    return callback_wrapper



"""The web-server configuration."""

import environ


@environ.config
class WebServerConfig:  # pylint: disable=too-few-public-methods
    """Config of a web-server."""

    host = environ.var('0.0.0.0', name='HOST')
    port = environ.var(8080, name='PORT', converter=int)



"""A very low-level GitHub API client."""

from asyncio import iscoroutinefunction
from typing import Any, Dict, Optional, Tuple

from gidgethub.aiohttp import GitHubAPI

# pylint: disable=relative-beyond-top-level
from .tokens import GitHubToken, GitHubOAuthToken, GitHubJWTToken
from .utils import accept_preview_version, mark_uninitialized_in_repr


@mark_uninitialized_in_repr
class RawGitHubAPI(GitHubAPI):
    """A low-level GitHub API client with a pre-populated token."""

    def __init__(
            self,
            token: GitHubToken,
            *,
            user_agent: Optional[str] = None,
            **kwargs: Any,
    ) -> None:
        """Initialize the GitHub client with token."""
        self._token = token
        kwargs.pop('oauth_token', None)
        kwargs.pop('jwt', None)
        super().__init__(
            requester=user_agent,
            **kwargs,
        )

    @property
    def is_initialized(self):
        """Return GitHub token presence."""
        return self._token is not None

    def __repr__(self):
        """Render a class instance representation."""
        cls_name = self.__class__.__name__
        init_args = (
            f'token={self._token!r}, '
            f'session={self._session!r}, '
            f'user_agent={self.requester!r}'
        )
        return f'{cls_name}({init_args})'

    # pylint: disable=arguments-differ
    # pylint: disable=keyword-arg-before-vararg
    # pylint: disable=too-many-arguments
    async def _make_request(
            self, method: str, url: str, url_vars: Dict[str, str],
            data: Any, accept: str = None,
            jwt: Optional[str] = None,
            oauth_token: Optional[str] = None,
    ) -> Tuple[bytes, Optional[str]]:
        token = self._token
        if iscoroutinefunction(token):
            token = await token()
        if isinstance(token, GitHubOAuthToken):
            oauth_token = str(token)
            jwt = None
        if isinstance(token, GitHubJWTToken):
            jwt = str(token)
            oauth_token = None
        return await super()._make_request(
            method=method,
            url=url,
            url_vars=url_vars,
            data=data,
            accept=accept,
            oauth_token=oauth_token,
            jwt=jwt,
        )

    getitem = accept_preview_version(GitHubAPI.getitem)
    getiter = accept_preview_version(GitHubAPI.getiter)
    post = accept_preview_version(GitHubAPI.post)
    patch = accept_preview_version(GitHubAPI.patch)
    put = accept_preview_version(GitHubAPI.put)
    delete = accept_preview_version(GitHubAPI.delete)



"""GitHub App runtime context helpers."""

from contextvars import ContextVar, Token
import logging
import os
import typing


logger = logging.getLogger(__name__)


class ContextLookupError(AttributeError):
    """Context var lookup error."""


class _ContextMap:
    __slots__ = '__map__', '__token_map__'

    def __init__(self, **initial_vars):
        self.__map__: typing.Dict[str, ContextVar[typing.Any]] = {
            k: ContextVar(v) for k, v in initial_vars.items()
        }
        """Storage for all context vars."""

        self.__token_map__: typing.Dict[str, Token[typing.Any]] = {}
        """Storage for individual context var reset tokens."""

    def __dir__(self):
        """Render a list of public attributes."""
        return self.__map__.keys()

    def __getattr__(self, name):
        if name in ('__map__', '__token_map__'):
            return getattr(self, name)
        try:
            return self.__map__[name].get()
        except LookupError:
            raise ContextLookupError(f'No `{name}` present in the context')

    def __setattr__(self, name, value):
        if name in ('__map__', '__token_map__'):
            object.__setattr__(self, name, value)
        elif name in self.__map__:
            reset_token = self.__map__[name].set(value)
            self.__token_map__[name] = reset_token
        else:
            raise ContextLookupError(f'No `{name}` present in the context')

    def __delattr__(self, name):
        if name not in self.__map__:
            raise ContextLookupError(f'No `{name}` present in the context')
        reset_token = self.__token_map__[name]
        self.__map__[name].reset(reset_token)
        del self.__token_map__[name]


def detect_env_mode():
    """Figure out whether we're under GitHub Action environment."""
    for var_suffix in {
            'WORKFLOW',
            'ACTION', 'ACTOR',
            'REPOSITORY',
            'EVENT_NAME', 'EVENT_PATH',
            'WORKSPACE',
            'SHA', 'REF',
            'TOKEN',
    }:
        if f'GITHUB_{var_suffix}' not in os.environ:
            logger.info(
                'Detected GitHub App mode since '
                'GITHUB_%s is missing from the env',
                var_suffix,
            )
            return 'app'
    logger.info(
        'Detected GitHub Action mode since all the '
        'typical env vars are present in the env',
    )
    return 'action'
